@@ -61,6 +61,9 @@ static int handle_connection(PRFileDesc *, PRFileDesc *, int);
 static const char envVarName[] = { SSL_ENV_VAR_NAME };
 static const char inheritableSockName[] = { "SELFSERV_LISTEN_SOCKET" };
 
+#define MAX_VIRT_SERVER_NAME_ARRAY_INDEX 10
+#define MAX_CERT_NICKNAME_ARRAY_INDEX 10
+
 #define DEFAULT_BULK_TEST 16384
 #define MAX_BULK_TEST 1048576 /* 1 MB */
 static PRBool testBulk;
@@ -92,35 +95,35 @@ static enum ocspStaplingModeEnum {
 } ocspStaplingMode = osm_disabled;
 typedef enum ocspStaplingModeEnum ocspStaplingModeType;
 static char *ocspStaplingCA = NULL;
-static SECItemArray *certStatus[kt_kea_size] = { NULL };
+static SECItemArray *certStatus[MAX_CERT_NICKNAME_ARRAY_INDEX] = { NULL };
 
 const int ssl3CipherSuites[] = {
-    -1,                                  /* SSL_FORTEZZA_DMS_WITH_FORTEZZA_CBC_SHA* a */
-    -1,                                  /* SSL_FORTEZZA_DMS_WITH_RC4_128_SHA    * b */
-    TLS_RSA_WITH_RC4_128_MD5,            /* c */
-    TLS_RSA_WITH_3DES_EDE_CBC_SHA,       /* d */
-    TLS_RSA_WITH_DES_CBC_SHA,            /* e */
-    TLS_RSA_EXPORT_WITH_RC4_40_MD5,      /* f */
-    TLS_RSA_EXPORT_WITH_RC2_CBC_40_MD5,  /* g */
-    -1,                                  /* SSL_FORTEZZA_DMS_WITH_NULL_SHA,  * h */
-    TLS_RSA_WITH_NULL_MD5,               /* i */
-    SSL_RSA_FIPS_WITH_3DES_EDE_CBC_SHA,  /* j */
-    SSL_RSA_FIPS_WITH_DES_CBC_SHA,       /* k */
-    TLS_RSA_EXPORT1024_WITH_DES_CBC_SHA, /* l */
-    TLS_RSA_EXPORT1024_WITH_RC4_56_SHA,  /* m */
-    TLS_RSA_WITH_RC4_128_SHA,            /* n */
-    TLS_DHE_DSS_WITH_RC4_128_SHA,        /* o */
-    TLS_DHE_RSA_WITH_3DES_EDE_CBC_SHA,   /* p */
-    TLS_DHE_DSS_WITH_3DES_EDE_CBC_SHA,   /* q */
-    TLS_DHE_RSA_WITH_DES_CBC_SHA,        /* r */
-    TLS_DHE_DSS_WITH_DES_CBC_SHA,        /* s */
-    TLS_DHE_DSS_WITH_AES_128_CBC_SHA,    /* t */
-    TLS_DHE_RSA_WITH_AES_128_CBC_SHA,    /* u */
-    TLS_RSA_WITH_AES_128_CBC_SHA,        /* v */
-    TLS_DHE_DSS_WITH_AES_256_CBC_SHA,    /* w */
-    TLS_DHE_RSA_WITH_AES_256_CBC_SHA,    /* x */
-    TLS_RSA_WITH_AES_256_CBC_SHA,        /* y */
-    TLS_RSA_WITH_NULL_SHA,               /* z */
+    -1,                                /* SSL_FORTEZZA_DMS_WITH_FORTEZZA_CBC_SHA* a */
+    -1,                                /* SSL_FORTEZZA_DMS_WITH_RC4_128_SHA     * b */
+    TLS_RSA_WITH_RC4_128_MD5,          /* c */
+    TLS_RSA_WITH_3DES_EDE_CBC_SHA,     /* d */
+    TLS_RSA_WITH_DES_CBC_SHA,          /* e */
+    -1,                                /* TLS_RSA_EXPORT_WITH_RC4_40_MD5        * f */
+    -1,                                /* TLS_RSA_EXPORT_WITH_RC2_CBC_40_MD5    * g */
+    -1,                                /* SSL_FORTEZZA_DMS_WITH_NULL_SHA        * h */
+    TLS_RSA_WITH_NULL_MD5,             /* i */
+    -1,                                /* SSL_RSA_FIPS_WITH_3DES_EDE_CBC_SHA    * j */
+    -1,                                /* SSL_RSA_FIPS_WITH_DES_CBC_SHA         * k */
+    -1,                                /* TLS_RSA_EXPORT1024_WITH_DES_CBC_SHA   * l */
+    -1,                                /* TLS_RSA_EXPORT1024_WITH_RC4_56_SHA    * m */
+    TLS_RSA_WITH_RC4_128_SHA,          /* n */
+    TLS_DHE_DSS_WITH_RC4_128_SHA,      /* o */
+    TLS_DHE_RSA_WITH_3DES_EDE_CBC_SHA, /* p */
+    TLS_DHE_DSS_WITH_3DES_EDE_CBC_SHA, /* q */
+    TLS_DHE_RSA_WITH_DES_CBC_SHA,      /* r */
+    TLS_DHE_DSS_WITH_DES_CBC_SHA,      /* s */
+    TLS_DHE_DSS_WITH_AES_128_CBC_SHA,  /* t */
+    TLS_DHE_RSA_WITH_AES_128_CBC_SHA,  /* u */
+    TLS_RSA_WITH_AES_128_CBC_SHA,      /* v */
+    TLS_DHE_DSS_WITH_AES_256_CBC_SHA,  /* w */
+    TLS_DHE_RSA_WITH_AES_256_CBC_SHA,  /* x */
+    TLS_RSA_WITH_AES_256_CBC_SHA,      /* y */
+    TLS_RSA_WITH_NULL_SHA,             /* z */
     0
 };
 
@@ -156,12 +159,12 @@ static void
 PrintUsageHeader(const char *progName)
 {
     fprintf(stderr,
-            "Usage: %s -n rsa_nickname -p port [-BDENRbjlmrsuvx] [-w password]\n"
+            "Usage: %s -n rsa_nickname -p port [-BDENRZbjlmrsuvx] [-w password]\n"
             "         [-t threads] [-i pid_file] [-c ciphers] [-Y] [-d dbdir] [-g numblocks]\n"
             "         [-f password_file] [-L [seconds]] [-M maxProcs] [-P dbprefix]\n"
             "         [-V [min-version]:[max-version]] [-a sni_name]\n"
             "         [ T <good|revoked|unknown|badsig|corrupted|none|ocsp>] [-A ca]\n"
-            "         [-C SSLCacheEntries] [-S dsa_nickname] -Q"
+            "         [-C SSLCacheEntries] [-S dsa_nickname] -Q [-I groups]"
 #ifndef NSS_DISABLE_ECC
             " [-e ec_nickname]"
 #endif /* NSS_DISABLE_ECC */
@@ -178,10 +181,7 @@ PrintParameterUsage()
         "   All versions are enabled by default.\n"
         "   Possible values for min/max: ssl3 tls1.0 tls1.1 tls1.2\n"
         "   Example: \"-V ssl3:\" enables SSL 3 and newer.\n"
-        "-B bypasses the PKCS11 layer for SSL encryption and MACing\n"
-        "-q checks for bypassability\n"
         "-D means disable Nagle delays in TCP\n"
-        "-E means disable export ciphersuites and SSL step down key gen\n"
         "-R means disable detection of rollback from TLS to SSL3\n"
         "-a configure server for SNI.\n"
         "-k expected name negotiated on server sockets\n"
@@ -195,7 +195,6 @@ PrintParameterUsage()
         "-s means disable SSL socket locking for performance\n"
         "-u means enable Session Ticket extension for TLS.\n"
         "-v means verbose output\n"
-        "-x means use export policy.\n"
         "-z means enable compression.\n"
         "-L seconds means log statistics every 'seconds' seconds (default=30).\n"
         "-M maxProcs tells how many processes to run in a multi-process server\n"
@@ -220,12 +219,16 @@ PrintParameterUsage()
         "-A <ca> Nickname of a CA used to sign a stapled cert status\n"
         "-U override default ECDHE ephemeral key reuse, 0: refresh, 1: reuse\n"
         "-H override default DHE server support, 0: disable, 1: enable, "
-        " 2: require DH named groups\n"
+        "   2: require DH named groups\n"
         "-W override default DHE server weak parameters support, 0: disable, 1: enable\n"
         "-c Restrict ciphers\n"
         "-Y prints cipher values allowed for parameter -c and exits\n"
         "-G enables the extended master secret extension [RFC7627]\n"
-        "-Q enables ALPN for HTTP/1.1 [RFC7301]\n",
+        "-Q enables ALPN for HTTP/1.1 [RFC7301]\n"
+        "-I comma separated list of enabled groups for TLS key exchange.\n"
+        "   The following values are valid:\n"
+        "   P256, P384, P521, x25519, FF2048, FF3072, FF4096, FF6144, FF8192\n"
+        "-Z enable 0-RTT (for TLS 1.3; also use -u)\n",
         stderr);
 }
 
@@ -313,48 +316,6 @@ disableAllSSLCiphers(void)
     }
 }
 
-/* disable all the export SSL cipher suites */
-SECStatus
-disableExportSSLCiphers(void)
-{
-    const PRUint16 *cipherSuites = SSL_ImplementedCiphers;
-    int i = SSL_NumImplementedCiphers;
-    SECStatus rv = SECSuccess;
-    SSLCipherSuiteInfo info;
-
-    while (--i >= 0) {
-        PRUint16 suite = cipherSuites[i];
-        SECStatus status;
-        status = SSL_GetCipherSuiteInfo(suite, &info, sizeof info);
-        if (status != SECSuccess) {
-            printf("SSL_GetCipherSuiteInfo rejected suite 0x%04x (i = %d)\n",
-                   suite, i);
-            errWarn("SSL_GetCipherSuiteInfo");
-            rv = SECFailure;
-            continue;
-        }
-        if (info.cipherSuite != suite) {
-            printf(
-                "SSL_GetCipherSuiteInfo returned wrong suite! Wanted 0x%04x, Got 0x%04x\n",
-                suite, i);
-            PORT_SetError(SEC_ERROR_LIBRARY_FAILURE);
-            rv = SECFailure;
-            continue;
-        }
-        /* should check here that info.length >= offsetof isExportable */
-        if (info.isExportable) {
-            status = SSL_CipherPolicySet(suite, SSL_NOT_ALLOWED);
-            if (status != SECSuccess) {
-                printf("SSL_CipherPolicySet rejected suite 0x%04x (i = %d)\n",
-                       suite, i);
-                errWarn("SSL_CipherPolicySet");
-                rv = SECFailure;
-            }
-        }
-    }
-    return rv;
-}
-
 static SECStatus
 mySSLAuthCertificate(void *arg, PRFileDesc *fd, PRBool checkSig,
                      PRBool isServer)
@@ -474,8 +435,6 @@ myBadCertHandler(void *arg, PRFileDesc *fd)
     return (MakeCertOK ? SECSuccess : SECFailure);
 }
 
-#define MAX_VIRT_SERVER_NAME_ARRAY_INDEX 10
-
 /* Simple SNI socket config function that does not use SSL_ReconfigFD.
  * Only uses one server name but verifies that the names match. */
 PRInt32
@@ -839,20 +798,22 @@ static SSLVersionRange enabledVersions;
 PRBool disableRollBack = PR_FALSE;
 PRBool NoReuse = PR_FALSE;
 PRBool hasSidCache = PR_FALSE;
-PRBool disableStepDown = PR_FALSE;
-PRBool bypassPKCS11 = PR_FALSE;
 PRBool disableLocking = PR_FALSE;
-PRBool testbypass = PR_FALSE;
 PRBool enableSessionTickets = PR_FALSE;
 PRBool enableCompression = PR_FALSE;
 PRBool failedToNegotiateName = PR_FALSE;
 PRBool enableExtendedMasterSecret = PR_FALSE;
 PRBool zeroRTT = PR_FALSE;
 PRBool enableALPN = PR_FALSE;
+SSLNamedGroup *enabledGroups = NULL;
+unsigned int enabledGroupsCount = 0;
 
 static char *virtServerNameArray[MAX_VIRT_SERVER_NAME_ARRAY_INDEX];
 static int virtServerNameIndex = 1;
 
+static char *certNicknameArray[MAX_CERT_NICKNAME_ARRAY_INDEX];
+static int certNicknameIndex = 0;
+
 static const char stopCmd[] = { "GET /stop " };
 static const char getCmd[] = { "GET " };
 static const char EOFmsg[] = { "EOF\r\n\r\n\r\n" };
@@ -1215,7 +1176,7 @@ makeSignedOCSPResponse(PLArenaPool *arena, ocspStaplingModeType osm,
 
 void
 setupCertStatus(PLArenaPool *arena, enum ocspStaplingModeEnum ocspStaplingMode,
-                CERTCertificate *cert, SSLKEAType kea, secuPWData *pwdata)
+                CERTCertificate *cert, int index, secuPWData *pwdata)
 {
     if (ocspStaplingMode == osm_random) {
         /* 6 different responses */
@@ -1269,7 +1230,7 @@ setupCertStatus(PLArenaPool *arena, enum ocspStaplingModeEnum ocspStaplingMode,
                 break;
         }
         if (multiOcspResponses) {
-            certStatus[kea] = multiOcspResponses;
+            certStatus[index] = multiOcspResponses;
         }
     }
 }
@@ -1296,11 +1257,10 @@ handle_connection(
     PRSocketOptionData opt;
     PRIOVec iovs[16];
     char msgBuf[160];
-    char buf[10240];
+    char buf[10240] = { 0 };
     char fileName[513];
     char proto[128];
     PRDescIdentity aboveLayer = PR_INVALID_IO_LAYER;
-    SSLKEAType kea;
 
     pBuf = buf;
     bufRem = sizeof buf;
@@ -1327,12 +1287,6 @@ handle_connection(
         ssl_sock = tcp_sock;
     }
 
-    for (kea = kt_rsa; kea < kt_kea_size; kea++) {
-        if (certStatus[kea] != NULL) {
-            SSL_SetStapledOCSPResponses(ssl_sock, certStatus[kea], kea);
-        }
-    }
-
     if (loggingLayer) {
         /* find the layer where our new layer is to be pushed */
         aboveLayer = PR_GetLayersIdentity(ssl_sock->lower);
@@ -1869,9 +1823,9 @@ server_main(
     CERTCertificate **cert,
     const char *expectedHostNameVal)
 {
+    int i;
     PRFileDesc *model_sock = NULL;
     int rv;
-    SSLKEAType kea;
     SECStatus secStatus;
 
     if (useModelSocket) {
@@ -1905,18 +1859,6 @@ server_main(
     if (rv != SECSuccess) {
         errExit("error enabling RollBack detection ");
     }
-    if (disableStepDown) {
-        rv = SSL_OptionSet(model_sock, SSL_NO_STEP_DOWN, PR_TRUE);
-        if (rv != SECSuccess) {
-            errExit("error disabling SSL StepDown ");
-        }
-    }
-    if (bypassPKCS11) {
-        rv = SSL_OptionSet(model_sock, SSL_BYPASS_PKCS11, PR_TRUE);
-        if (rv != SECSuccess) {
-            errExit("error enabling PKCS11 bypass ");
-        }
-    }
     if (disableLocking) {
         rv = SSL_OptionSet(model_sock, SSL_NO_LOCKS, PR_TRUE);
         if (rv != SECSuccess) {
@@ -1980,12 +1922,17 @@ server_main(
 
     /* This uses the legacy certificate API.  See mySSLSNISocketConfig() for the
      * new, prefered API. */
-    for (kea = kt_rsa; kea < kt_kea_size; kea++) {
-        if (cert[kea] != NULL) {
-            secStatus = SSL_ConfigSecureServer(model_sock,
-                                               cert[kea], privKey[kea], kea);
+    for (i = 0; i < certNicknameIndex; i++) {
+        if (cert[i] != NULL) {
+            const SSLExtraServerCertData ocspData = {
+                ssl_auth_null, NULL, certStatus[i], NULL
+            };
+
+            secStatus = SSL_ConfigServerCert(model_sock, cert[i],
+                                             privKey[i], &ocspData,
+                                             sizeof(ocspData));
             if (secStatus != SECSuccess)
-                errExit("SSL_ConfigSecureServer");
+                errExit("SSL_ConfigServerCert");
         }
     }
 
@@ -2027,6 +1974,13 @@ server_main(
         }
     }
 
+    if (enabledGroups) {
+        rv = SSL_NamedGroupConfig(model_sock, enabledGroups, enabledGroupsCount);
+        if (rv < 0) {
+            errExit("SSL_NamedGroupConfig failed");
+        }
+    }
+
     /* This cipher is not on by default. The Acceptance test
      * would like it to be. Turn this cipher on.
      */
@@ -2230,11 +2184,6 @@ int
 main(int argc, char **argv)
 {
     char *progName = NULL;
-    char *nickName = NULL;
-#ifndef NSS_DISABLE_ECC
-    char *ecNickName = NULL;
-#endif
-    char *dsaNickName = NULL;
     const char *fileName = NULL;
     char *cipherString = NULL;
     const char *dir = ".";
@@ -2244,15 +2193,14 @@ main(int argc, char **argv)
     char *tmp;
     char *envString;
     PRFileDesc *listen_sock;
-    CERTCertificate *cert[kt_kea_size] = { NULL };
-    SECKEYPrivateKey *privKey[kt_kea_size] = { NULL };
+    CERTCertificate *cert[MAX_CERT_NICKNAME_ARRAY_INDEX] = { NULL };
+    SECKEYPrivateKey *privKey[MAX_CERT_NICKNAME_ARRAY_INDEX] = { NULL };
     int optionsFound = 0;
     int maxProcs = 1;
     unsigned short port = 0;
-    SECStatus rv;
+    SECStatus rv = SECSuccess;
     PRStatus prStatus;
     PRBool bindOnly = PR_FALSE;
-    PRBool useExportPolicy = PR_FALSE;
     PRBool useLocalThreads = PR_FALSE;
     PLOptState *optstate;
     PLOptStatus status;
@@ -2260,7 +2208,6 @@ main(int argc, char **argv)
     PRBool debugCache = PR_FALSE; /* bug 90518 */
     char emptyString[] = { "" };
     char *certPrefix = emptyString;
-    PRUint32 protos = 0;
     SSL3Statistics *ssl3stats;
     PRUint32 i;
     secuPWData pwdata = { PW_NONE, 0 };
@@ -2277,9 +2224,10 @@ main(int argc, char **argv)
 
     /* please keep this list of options in ASCII collating sequence.
     ** numbers, then capital letters, then lower case, alphabetical.
-    */
+    ** XXX: 'B', 'E', 'q', and 'x' were used in the past but removed
+    **      in 3.28, please leave some time before resuing those. */
     optstate = PL_CreateOptState(argc, argv,
-                                 "2:A:BC:DEGH:L:M:NP:QRS:T:U:V:W:YZa:bc:d:e:f:g:hi:jk:lmn:op:qrst:uvw:xyz");
+                                 "2:A:C:DGH:I:L:M:NP:QRS:T:U:V:W:YZa:bc:d:e:f:g:hi:jk:lmn:op:rst:uvw:yz");
     while ((status = PL_GetNextOpt(optstate)) == PL_OPT_OK) {
         ++optionsFound;
         switch (optstate->option) {
@@ -2291,10 +2239,6 @@ main(int argc, char **argv)
                 ocspStaplingCA = PORT_Strdup(optstate->value);
                 break;
 
-            case 'B':
-                bypassPKCS11 = PR_TRUE;
-                break;
-
             case 'C':
                 if (optstate->value)
                     NumSidCacheEntries = PORT_Atoi(optstate->value);
@@ -2303,9 +2247,6 @@ main(int argc, char **argv)
             case 'D':
                 noDelay = PR_TRUE;
                 break;
-            case 'E':
-                disableStepDown = PR_TRUE;
-                break;
             case 'H':
                 configureDHE = (PORT_Atoi(optstate->value) != 0);
                 break;
@@ -2314,9 +2255,6 @@ main(int argc, char **argv)
                 enableExtendedMasterSecret = PR_TRUE;
                 break;
 
-            case 'I': /* reserved for OCSP multi-stapling */
-                break;
-
             case 'L':
                 logStats = PR_TRUE;
                 if (optstate->value == NULL) {
@@ -2345,7 +2283,11 @@ main(int argc, char **argv)
                 break;
 
             case 'S':
-                dsaNickName = PORT_Strdup(optstate->value);
+                if (certNicknameIndex >= MAX_CERT_NICKNAME_ARRAY_INDEX) {
+                    Usage(progName);
+                    break;
+                }
+                certNicknameArray[certNicknameIndex++] = PORT_Strdup(optstate->value);
                 break;
 
             case 'T':
@@ -2364,7 +2306,9 @@ main(int argc, char **argv)
                 if (SECU_ParseSSLVersionRangeString(optstate->value,
                                                     enabledVersions, &enabledVersions) !=
                     SECSuccess) {
+                    fprintf(stderr, "Bad version specified.\n");
                     Usage(progName);
+                    exit(1);
                 }
                 break;
 
@@ -2400,7 +2344,11 @@ main(int argc, char **argv)
 
 #ifndef NSS_DISABLE_ECC
             case 'e':
-                ecNickName = PORT_Strdup(optstate->value);
+                if (certNicknameIndex >= MAX_CERT_NICKNAME_ARRAY_INDEX) {
+                    Usage(progName);
+                    break;
+                }
+                certNicknameArray[certNicknameIndex++] = PORT_Strdup(optstate->value);
                 break;
 #endif /* NSS_DISABLE_ECC */
 
@@ -2441,7 +2389,11 @@ main(int argc, char **argv)
                 break;
 
             case 'n':
-                nickName = PORT_Strdup(optstate->value);
+                if (certNicknameIndex >= MAX_CERT_NICKNAME_ARRAY_INDEX) {
+                    Usage(progName);
+                    break;
+                }
+                certNicknameArray[certNicknameIndex++] = PORT_Strdup(optstate->value);
                 virtServerNameArray[0] = PORT_Strdup(optstate->value);
                 break;
 
@@ -2457,10 +2409,6 @@ main(int argc, char **argv)
                 port = PORT_Atoi(optstate->value);
                 break;
 
-            case 'q':
-                testbypass = PR_TRUE;
-                break;
-
             case 'r':
                 ++requestCert;
                 break;
@@ -2490,10 +2438,6 @@ main(int argc, char **argv)
                 pwdata.data = passwd = PORT_Strdup(optstate->value);
                 break;
 
-            case 'x':
-                useExportPolicy = PR_TRUE;
-                break;
-
             case 'y':
                 debugCache = PR_TRUE;
                 break;
@@ -2510,6 +2454,16 @@ main(int argc, char **argv)
                 enableALPN = PR_TRUE;
                 break;
 
+            case 'I':
+                rv = parseGroupList(optstate->value, &enabledGroups, &enabledGroupsCount);
+                if (rv != SECSuccess) {
+                    PL_DestroyOptState(optstate);
+                    fprintf(stderr, "Bad group specified.\n");
+                    fprintf(stderr, "Run '%s -h' for usage information.\n", progName);
+                    exit(5);
+                }
+                break;
+
             default:
             case '?':
                 fprintf(stderr, "Unrecognized or bad option specified.\n");
@@ -2558,14 +2512,8 @@ main(int argc, char **argv)
         exit(0);
     }
 
-    if ((nickName == NULL) &&
-        (dsaNickName == NULL)
-#ifndef NSS_DISABLE_ECC
-        && (ecNickName == NULL)
-#endif
-            ) {
-
-        fprintf(stderr, "Required arg '-n' (rsa nickname) not supplied.\n");
+    if (certNicknameIndex == 0) {
+        fprintf(stderr, "Must specify at least one certificate nickname using '-n' (RSA), '-S' (DSA), or 'e' (EC).\n");
         fprintf(stderr, "Run '%s -h' for usage information.\n", progName);
         exit(6);
     }
@@ -2665,24 +2613,6 @@ main(int argc, char **argv)
         exit(8);
     }
 
-    /* set the policy bits true for all the cipher suites. */
-    if (useExportPolicy) {
-        NSS_SetExportPolicy();
-        if (disableStepDown) {
-            fputs("selfserv: -x and -E options may not be used together\n",
-                  stderr);
-            exit(98);
-        }
-    } else {
-        NSS_SetDomesticPolicy();
-        if (disableStepDown) {
-            rv = disableExportSSLCiphers();
-            if (rv != SECSuccess) {
-                errExit("error disabling export ciphersuites ");
-            }
-        }
-    }
-
     /* all SSL3 cipher suites are enabled by default. */
     if (cipherString) {
         char *cstringSaved = cipherString;
@@ -2734,123 +2664,34 @@ main(int argc, char **argv)
         PORT_Free(cstringSaved);
     }
 
-    if (testbypass) {
-        const PRUint16 *cipherSuites = SSL_ImplementedCiphers;
-        int i = SSL_NumImplementedCiphers;
-        PRBool enabled;
-
-        for (i = 0; i < SSL_NumImplementedCiphers; i++, cipherSuites++) {
-            if (SSL_CipherPrefGetDefault(*cipherSuites, &enabled) == SECSuccess &&
-                enabled)
-                savecipher(*cipherSuites);
-        }
-        protos = 0;
-        if (enabledVersions.min <= SSL_LIBRARY_VERSION_3_0 &&
-            enabledVersions.max >= SSL_LIBRARY_VERSION_3_0) {
-            protos |= SSL_CBP_SSL3;
-        }
-        if (enabledVersions.min <= SSL_LIBRARY_VERSION_TLS_1_0 &&
-            enabledVersions.max >= SSL_LIBRARY_VERSION_TLS_1_0) {
-            protos |= SSL_CBP_TLS1_0;
-        }
-        /* TLS 1.1 has the same SSL Bypass mode requirements as TLS 1.0 */
-        if (enabledVersions.min <= SSL_LIBRARY_VERSION_TLS_1_1 &&
-            enabledVersions.max >= SSL_LIBRARY_VERSION_TLS_1_1) {
-            protos |= SSL_CBP_TLS1_0;
-        }
-    }
-
     certStatusArena = PORT_NewArena(DER_DEFAULT_CHUNKSIZE);
     if (!certStatusArena)
         errExit("cannot allocate certStatusArena");
 
-    if (nickName) {
-        cert[kt_rsa] = PK11_FindCertFromNickname(nickName, &pwdata);
-        if (cert[kt_rsa] == NULL) {
-            fprintf(stderr, "selfserv: Can't find certificate %s\n", nickName);
+    for (i = 0; i < certNicknameIndex; i++) {
+        cert[i] = PK11_FindCertFromNickname(certNicknameArray[i], &pwdata);
+        if (cert[i] == NULL) {
+            fprintf(stderr, "selfserv: Can't find certificate %s\n", certNicknameArray[i]);
             exit(10);
         }
-        privKey[kt_rsa] = PK11_FindKeyByAnyCert(cert[kt_rsa], &pwdata);
-        if (privKey[kt_rsa] == NULL) {
+        privKey[i] = PK11_FindKeyByAnyCert(cert[i], &pwdata);
+        if (privKey[i] == NULL) {
             fprintf(stderr, "selfserv: Can't find Private Key for cert %s\n",
-                    nickName);
+                    certNicknameArray[i]);
             exit(11);
         }
-        if (testbypass) {
-            PRBool bypassOK;
-            if (SSL_CanBypass(cert[kt_rsa], privKey[kt_rsa], protos, cipherlist,
-                              nciphers, &bypassOK, &pwdata) != SECSuccess) {
-                SECU_PrintError(progName, "Bypass test failed %s\n", nickName);
-                exit(14);
-            }
-            fprintf(stderr, "selfserv: %s can%s bypass\n", nickName,
-                    bypassOK ? "" : "not");
-        }
-        setupCertStatus(certStatusArena, ocspStaplingMode, cert[kt_rsa], kt_rsa,
-                        &pwdata);
+#ifdef NSS_DISABLE_ECC
+        if (privKey[i]->keyType != ecKey)
+#endif
+            setupCertStatus(certStatusArena, ocspStaplingMode, cert[i], i, &pwdata);
     }
-    if (dsaNickName) {
-        /* Investigate if ssl_kea_dh should be changed to ssl_auth_dsa.
-         * See bug 102794.*/
-        cert[ssl_kea_dh] = PK11_FindCertFromNickname(dsaNickName, &pwdata);
-        if (cert[ssl_kea_dh] == NULL) {
-            fprintf(stderr, "selfserv: Can't find certificate %s\n", dsaNickName);
-            exit(12);
-        }
-        privKey[ssl_kea_dh] = PK11_FindKeyByAnyCert(cert[ssl_kea_dh], &pwdata);
-        if (privKey[ssl_kea_dh] == NULL) {
-            fprintf(stderr, "selfserv: Can't find Private Key for cert %s\n",
-                    dsaNickName);
-            exit(11);
-        }
-        if (testbypass) {
-            PRBool bypassOK;
-            if (SSL_CanBypass(cert[ssl_kea_dh], privKey[ssl_kea_dh], protos, cipherlist,
-                              nciphers, &bypassOK, &pwdata) != SECSuccess) {
-                SECU_PrintError(progName, "Bypass test failed %s\n", nickName);
-                exit(14);
-            }
-            fprintf(stderr, "selfserv: %s can%s bypass\n", nickName,
-                    bypassOK ? "" : "not");
-        }
-        setupCertStatus(certStatusArena, ocspStaplingMode, cert[ssl_kea_dh], ssl_kea_dh,
-                        &pwdata);
-    }
-#ifndef NSS_DISABLE_ECC
-    if (ecNickName) {
-        cert[kt_ecdh] = PK11_FindCertFromNickname(ecNickName, &pwdata);
-        if (cert[kt_ecdh] == NULL) {
-            fprintf(stderr, "selfserv: Can't find certificate %s\n",
-                    ecNickName);
-            exit(13);
-        }
-        privKey[kt_ecdh] = PK11_FindKeyByAnyCert(cert[kt_ecdh], &pwdata);
-        if (privKey[kt_ecdh] == NULL) {
-            fprintf(stderr, "selfserv: Can't find Private Key for cert %s\n",
-                    ecNickName);
-            exit(11);
-        }
-        if (testbypass) {
-            PRBool bypassOK;
-            if (SSL_CanBypass(cert[kt_ecdh], privKey[kt_ecdh], protos, cipherlist,
-                              nciphers, &bypassOK, &pwdata) != SECSuccess) {
-                SECU_PrintError(progName, "Bypass test failed %s\n", ecNickName);
-                exit(15);
-            }
-            fprintf(stderr, "selfserv: %s can%s bypass\n", ecNickName,
-                    bypassOK ? "" : "not");
-        }
-        setupCertStatus(certStatusArena, ocspStaplingMode, cert[kt_ecdh], kt_ecdh,
-                        &pwdata);
-    }
-#endif /* NSS_DISABLE_ECC */
-
-    if (testbypass)
-        goto cleanup;
 
     if (configureWeakDHE > 0) {
         fprintf(stderr, "selfserv: Creating dynamic weak DH parameters\n");
         rv = SSL_EnableWeakDHEPrimeGroup(NULL, PR_TRUE);
+        if (rv != SECSuccess) {
+            goto cleanup;
+        }
         fprintf(stderr, "selfserv: Done creating dynamic weak DH parameters\n");
     }
 
@@ -2890,13 +2731,14 @@ cleanup:
 
     {
         int i;
-        for (i = 0; i < kt_kea_size; i++) {
+        for (i = 0; i < certNicknameIndex; i++) {
             if (cert[i]) {
                 CERT_DestroyCertificate(cert[i]);
             }
             if (privKey[i]) {
                 SECKEY_DestroyPrivateKey(privKey[i]);
             }
+            PORT_Free(certNicknameArray[i]);
         }
         for (i = 0; virtServerNameArray[i]; i++) {
             PORT_Free(virtServerNameArray[i]);
@@ -2906,9 +2748,6 @@ cleanup:
     if (debugCache) {
         nss_DumpCertificateCacheInfo();
     }
-    if (nickName) {
-        PORT_Free(nickName);
-    }
     if (expectedHostNameVal) {
         PORT_Free(expectedHostNameVal);
     }
@@ -2921,14 +2760,6 @@ cleanup:
     if (certPrefix && certPrefix != emptyString) {
         PORT_Free(certPrefix);
     }
-#ifndef NSS_DISABLE_ECC
-    if (ecNickName) {
-        PORT_Free(ecNickName);
-    }
-#endif
-    if (dsaNickName) {
-        PORT_Free(dsaNickName);
-    }
 
     if (hasSidCache) {
         SSL_ShutdownServerSessionIDCache();
@@ -2936,6 +2767,9 @@ cleanup:
     if (certStatusArena) {
         PORT_FreeArena(certStatusArena, PR_FALSE);
     }
+    if (enabledGroups) {
+        PORT_Free(enabledGroups);
+    }
     if (NSS_Shutdown() != SECSuccess) {
         SECU_PrintError(progName, "NSS_Shutdown");
         if (loggerThread) {
